@@ -2,6 +2,7 @@ use std::net::IpAddr;
 use std::str::FromStr;
 
 use serde::Deserialize;
+use serde_json::json;
 
 use super::{MaxMindDBError, Reader};
 
@@ -41,7 +42,7 @@ fn test_decoder() {
 
     let r = Reader::open_readfile("test-data/test-data/MaxMind-DB-test-decoder.mmdb");
     if let Err(err) = r {
-        panic!("error opening mmdb: {:?}", err);
+        panic!("error opening mmdb: {err:?}");
     }
     let r = r.unwrap();
     let ip: IpAddr = FromStr::from_str("1.1.1.0").unwrap();
@@ -84,7 +85,7 @@ fn test_pointers_in_metadata() {
 
     let r = Reader::open_readfile("test-data/test-data/MaxMind-DB-test-metadata-pointers.mmdb");
     if let Err(err) = r {
-        panic!("error opening mmdb: {:?}", err);
+        panic!("error opening mmdb: {err:?}");
     }
 }
 
@@ -149,10 +150,8 @@ fn test_reader() {
     for record_size in &sizes {
         let versions = [4_usize, 6];
         for ip_version in &versions {
-            let filename = format!(
-                "test-data/test-data/MaxMind-DB-test-ipv{}-{}.mmdb",
-                ip_version, record_size
-            );
+            let filename =
+                format!("test-data/test-data/MaxMind-DB-test-ipv{ip_version}-{record_size}.mmdb");
             let reader = Reader::open_readfile(filename).ok().unwrap();
 
             check_metadata(&reader, *ip_version, *record_size);
@@ -170,10 +169,8 @@ fn test_reader_readfile() {
     for record_size in &sizes {
         let versions = [4_usize, 6];
         for ip_version in &versions {
-            let filename = format!(
-                "test-data/test-data/MaxMind-DB-test-ipv{}-{}.mmdb",
-                ip_version, record_size
-            );
+            let filename =
+                format!("test-data/test-data/MaxMind-DB-test-ipv{ip_version}-{record_size}.mmdb");
             let reader = Reader::open_readfile(filename).ok().unwrap();
 
             check_metadata(&reader, *ip_version, *record_size);
@@ -382,7 +379,7 @@ fn test_within_city() {
     // Make sure the first is what we expect it to be
     let item = iter.next().unwrap().unwrap();
     assert_eq!(item.ip_net, IpNetwork::V4("2.2.3.0/24".parse().unwrap()));
-    assert_eq!(item.info.city.unwrap().geoname_id, Some(2655045));
+    assert_eq!(item.info.city.unwrap().geoname_id, Some(2_655_045));
 
     let mut n = 1;
     for _ in iter {
@@ -403,8 +400,7 @@ fn test_within_city() {
         IpNetwork::V4("81.2.69.144/28".parse().unwrap()),
         IpNetwork::V4("81.2.69.142/31".parse().unwrap()),
     ];
-    while !expected.is_empty() {
-        let e = expected.pop().unwrap();
+    while let Some(e) = expected.pop() {
         let item = iter.next().unwrap().unwrap();
         assert_eq!(item.ip_net, e);
     }
@@ -480,7 +476,7 @@ fn check_ip<T: AsRef<[u8]>>(reader: &Reader<T>, ip_version: usize) {
     for &address in &no_record {
         let ip: IpAddr = FromStr::from_str(address).unwrap();
         match reader.lookup::<IpType>(ip) {
-            Ok(v) => panic!("received an unexpected value: {:?}", v),
+            Ok(v) => panic!("received an unexpected value: {v:?}"),
             Err(e) => assert_eq!(
                 e,
                 MaxMindDBError::AddressNotFoundError("Address not found in database".to_string())
@@ -488,3 +484,23 @@ fn check_ip<T: AsRef<[u8]>>(reader: &Reader<T>, ip_version: usize) {
         }
     }
 }
+
+#[test]
+fn test_json_serialize() {
+    use super::geoip2::City;
+    let _ = env_logger::try_init();
+
+    let filename = "test-data/test-data/GeoIP2-City-Test.mmdb";
+
+    let reader = Reader::open_readfile(filename).unwrap();
+
+    let ip: IpAddr = FromStr::from_str("89.160.20.112").unwrap();
+    let city: City = reader.lookup(ip).unwrap();
+
+    let json_string = json!(city).to_string();
+
+    assert_eq!(
+        json_string,
+        r#"{"city":{"geoname_id":2694762,"names":{"de":"Linköping","en":"Linköping","fr":"Linköping","ja":"リンシェーピング","zh-CN":"林雪平"}},"continent":{"code":"EU","geoname_id":6255148,"names":{"de":"Europa","en":"Europe","es":"Europa","fr":"Europe","ja":"ヨーロッパ","pt-BR":"Europa","ru":"Европа","zh-CN":"欧洲"}},"country":{"geoname_id":2661886,"is_in_european_union":true,"iso_code":"SE","names":{"de":"Schweden","en":"Sweden","es":"Suecia","fr":"Suède","ja":"スウェーデン王国","pt-BR":"Suécia","ru":"Швеция","zh-CN":"瑞典"}},"location":{"accuracy_radius":76,"latitude":58.4167,"longitude":15.6167,"time_zone":"Europe/Stockholm"},"registered_country":{"geoname_id":2921044,"is_in_european_union":true,"iso_code":"DE","names":{"de":"Deutschland","en":"Germany","es":"Alemania","fr":"Allemagne","ja":"ドイツ連邦共和国","pt-BR":"Alemanha","ru":"Германия","zh-CN":"德国"}},"subdivisions":[{"geoname_id":2685867,"iso_code":"E","names":{"en":"Östergötland County","fr":"Comté d'Östergötland"}}]}"#
+    );
+}
